package ch.psi.daq.queryrest.config;

import java.util.LinkedHashSet;
import java.util.Set;

import javax.annotation.Resource;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Import;
import org.springframework.context.annotation.PropertySource;
import org.springframework.core.env.Environment;
import org.springframework.util.StringUtils;

import com.fasterxml.jackson.annotation.JsonInclude.Include;
import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.Version;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.module.SimpleModule;

import ch.psi.daq.cassandra.util.test.CassandraDataGen;
import ch.psi.daq.common.json.deserialize.AttributeBasedDeserializer;
import ch.psi.daq.common.statistic.StorelessStatistics;
import ch.psi.daq.domain.cassandra.ChannelEvent;
import ch.psi.daq.query.config.QueryConfig;
import ch.psi.daq.query.model.AbstractQuery;
import ch.psi.daq.query.model.QueryField;
import ch.psi.daq.queryrest.model.PropertyFilterMixin;
import ch.psi.daq.queryrest.response.ResponseStreamWriter;

@Configuration
@PropertySource(value = {"classpath:queryrest.properties"})
@PropertySource(value = {"file:${user.home}/.config/daq/queryrest.properties"}, ignoreResourceNotFound = true)
public class QueryRestConfig {

   // a nested configuration
   // this guarantees that the ordering of the properties file is as expected
   // see:
   // https://jira.spring.io/browse/SPR-10409?focusedCommentId=101393&page=com.atlassian.jira.plugin.system.issuetabpanels:comment-tabpanel#comment-101393
   @Configuration
   @Import({QueryConfig.class})
   static class InnerConfiguration {
   }

   private static final Logger LOGGER = LoggerFactory.getLogger(QueryRestConfig.class);

   public static final String BEAN_NAME_DEFAULT_RESPONSE_FIELDS = "defaultResponseFields";

   @Resource
   private Environment env;

   @Bean
   public ObjectMapper objectMapper() {
      ObjectMapper mapper = new ObjectMapper();

      String packageName = AbstractQuery.class.getPackage().getName();
      Class<AbstractQuery> abstractQueryClass = AbstractQuery.class;
      AttributeBasedDeserializer<AbstractQuery> abstractQueryDeserializer =
            new AttributeBasedDeserializer<AbstractQuery>(abstractQueryClass).register(packageName);
      SimpleModule module = new SimpleModule("PolymorphicAbstractQuery", Version.unknownVersion());
      module.addDeserializer(abstractQueryClass, abstractQueryDeserializer);
      mapper.registerModule(module);

      // only include non-null values
      mapper.setSerializationInclusion(Include.NON_NULL);
      // Mixin which is used dynamically to filter out which properties get serialised and which
      // won't. This way, the user can specify which columns are to be received.
      mapper.addMixIn(ChannelEvent.class, PropertyFilterMixin.class);
      mapper.addMixIn(StorelessStatistics.class, PropertyFilterMixin.class);

      return mapper;
   }

   @Bean
   public JsonFactory jsonFactory() {
      return new JsonFactory();
   }

   @Bean
   public ResponseStreamWriter responseStreamWriter() {
      return new ResponseStreamWriter();
   }

   @Bean(name = BEAN_NAME_DEFAULT_RESPONSE_FIELDS)
   public Set<QueryField> defaultResponseFields() {
      String[] responseFields =
            StringUtils.commaDelimitedListToStringArray(env.getProperty("queryrest.default.response.fields"));
      // preserve order
      LinkedHashSet<QueryField> defaultResponseFields = new LinkedHashSet<>(responseFields.length);
      for (String field : responseFields) {
         try {
            defaultResponseFields.add(QueryField.valueOf(field));
         } catch (Exception e) {
            LOGGER.error("Field '{}' in queryrest.default.response.fields is invalid.", field, e);
         }
      }

      return defaultResponseFields;
   }
   
   
   // ==========================================================================================
   // TODO: This is simply for initial / rudimentary testing - remove once further evolved
   @Bean
   public CassandraDataGen cassandraDataGen() {
      return new CassandraDataGen();
   }
}
